import os
import numpy as np
import netCDF4 as nc
from collections import namedtuple

# Define the attributes of a netCDF variable
NetCDFVariable = namedtuple('NetCDFVariable', ['nickname', 'name', 'unit'])


class Ncio(object):
    def __init__(self, param):
        for key, val in param.items():
            setattr(self, key, val)

        self.experiment_parameters = param
        full_hist_variables = [NetCDFVariable('u', 'u', 'm/s'),
                               NetCDFVariable('v', 'v', 'm/s'),
                               NetCDFVariable('h', 'h', 'm'),
                               NetCDFVariable('pv', 'pv', '1/s')]

        self.hist_variables = full_hist_variables
        self.diag_variables = [NetCDFVariable('hm', 'mean height', 'm'),
                               NetCDFVariable('ke', 'kinetic', 'm^2/s^2'),
                               NetCDFVariable('pe', 'potential', 'm^2/s^2'),
                               NetCDFVariable('pvm', 'mean total pv', '1/s'),
                               NetCDFVariable('ens', 'enstrophy', '1/s^2')]

        dimensions = {'i': self.nx, 'j': self.ny}

        self.hist_path = 'history.nc'
        self.create_history_file(dimensions)
        self.kt = 0
        self.ktdiag = 0

    def create_history_file(self, dimensions):
        with nc.Dataset(self.hist_path, "w", format='NETCDF4') as ncfile:
            # Store the experiment parameters
            ncfile.setncatts(self.experiment_parameters)

            # Create the dimensions
            ncfile.createDimension("t", None)  # unlimited size
            ncfile.createDimension("tdiag", None)  # unlimited size
            ncfile.createDimension("x", dimensions["i"])
            ncfile.createDimension("y", dimensions["j"])

            # Create the variables with one dimension
            v = ncfile.createVariable("n", int, ("t",))
            v.long_name = "integration step in the model run"

            # TODO: make the units of x,y,z,t more general
            v = ncfile.createVariable("tdiag", float, ("tdiag",))
            v.long_name = "time in the model run"
            v.units = "s"

            # TODO: make the units of x,y,z,t more general
            v = ncfile.createVariable("t", float, ("t",))
            v.long_name = "time in the model run"
            v.units = "s"

            v = ncfile.createVariable("x", float, ("x",))
            v.long_name = "coordinate in the x-direction"
            v.units = "m"

            v = ncfile.createVariable("y", float, ("y",))
            v.long_name = "coordinate in the y-direction"
            v.units = "m"

            for variable in self.hist_variables:
                # Spatial dimensions are in reversed order, because
                # the arrays are stored like this in the Scalar class
                v = ncfile.createVariable(
                    variable.nickname, float, ("t", "y", "x"))
                v.long_name = variable.name
                v.units = variable.unit

            for variable in self.diag_variables:
                # Spatial dimensions are in reversed order, because
                # the arrays are stored like this in the Scalar class
                v = ncfile.createVariable(
                    variable.nickname, float, ("tdiag"))
                v.long_name = variable.name
                v.units = variable.unit

    def dohis(self, state, time):
        h, u, v, pv = state
        with nc.Dataset(self.hist_path, "r+") as ncfile:
            ncfile.variables['n'][self.kt] = self.kt
            ncfile.variables['t'][self.kt] = time
            ncfile.variables['h'][self.kt, :, :] = h
            ncfile.variables['u'][self.kt, :, :] = u[:, :-1]
            ncfile.variables['v'][self.kt, :, :] = v[:-1, :]
            ncfile.variables['pv'][self.kt, :, :] = pv
        self.kt += 1

    def dodiag(self, diags, time):
        with nc.Dataset(self.hist_path, "r+") as ncfile:
            ncfile.variables['tdiag'][self.ktdiag] = time
            for key, val in diags.items():
                ncfile.variables[key][self.ktdiag] = val
        self.ktdiag += 1


if __name__ == '__main__':
    nx, ny = 40, 50
    io = Ncio({'nx': nx, 'ny': ny})
    h = np.random.uniform(size=(ny, nx))
    u = np.random.uniform(size=(ny, nx+1))
    v = np.random.uniform(size=(ny+1, nx))
    pv = np.random.uniform(size=(ny, nx))
    state = [h, u, v, pv]
    t = 0.
    for kt in range(10):
        h[:, :] = np.random.uniform(size=(ny, nx))
        u[:, :] = np.random.uniform(size=(ny, nx+1))
        v[:, :] = np.random.uniform(size=(ny+1, nx))
        pv[:, :] = np.random.uniform(size=(ny, nx))
        io.dohis(state, t)
        t += 0.1
